import time
import numpy as np
import cv2
import signal
from threading import Thread
import subprocess

from shufflecad.shufflecad import Shufflecad
from shufflecad.shared import InfoHolder


class RobocadVMXTitan:
    def __init__(self, is_real_robot: bool = True):
        self.__motor_speed_0 = 0.0
        self.__motor_speed_1 = 0.0
        self.__motor_speed_2 = 0.0
        self.__motor_speed_3 = 0.0

        self.__motor_enc_0 = 0.0
        self.__motor_enc_1 = 0.0
        self.__motor_enc_2 = 0.0
        self.__motor_enc_3 = 0.0

        self.__limit_l_0 = False
        self.__limit_h_0 = False
        self.__limit_l_1 = False
        self.__limit_h_1 = False
        self.__limit_l_2 = False
        self.__limit_h_2 = False
        self.__limit_l_3 = False
        self.__limit_h_3 = False

        self.__yaw = 0.0
        # self.__yaw_unlim = 0.0

        self.__ultrasound_1 = 0.0
        self.__ultrasound_2 = 0.0

        self.__analog_1 = 0.0
        self.__analog_2 = 0.0
        self.__analog_3 = 0.0
        self.__analog_4 = 0.0

        self.__flex_0 = False
        self.__flex_1 = False
        self.__flex_2 = False
        self.__flex_3 = False
        self.__flex_4 = False
        self.__flex_5 = False
        self.__flex_6 = False
        self.__flex_7 = False

        self.__hcdio_values: list = [0] * 10

        self.__camera_image = None

        self.is_real_robot = is_real_robot
        InfoHolder.on_real_robot = is_real_robot
        Shufflecad.start()

        signal.signal(signal.SIGTERM, self.handler)
        signal.signal(signal.SIGINT, self.handler)

        if not self.is_real_robot:
            from .robocadSim.connection_helper import ConnectionHelper
            self.__connection_helper = ConnectionHelper(ConnectionHelper.CONN_ALL)
            self.__connection_helper.start_channels()

            InfoHolder.power = "12"  # :)
        else:
            try:
                self.__camera_instance = cv2.VideoCapture(0)
            except Exception as e:
                InfoHolder.logger.write_main_log("Exception while creating camera instance: ")
                InfoHolder.logger.write_main_log(str(e))
            # mb cringe
            global VMXStatic, TitanStatic
            from .pycad.shared import VMXStatic, TitanStatic
            from .pycad.SPI import VMXSPI
            from .pycad.COM import TitanCOM
            VMXSPI.start_spi()
            TitanCOM.start_com()
            subprocess.run(['sudo', '/home/pi/pi-blaster/pi-blaster'])
            th: Thread = Thread(target=RobocadVMXTitan.__update_rpi_cringe)
            th.daemon = True
            th.start()

    @classmethod
    def __update_rpi_cringe(cls):
        from gpiozero import CPUTemperature
        import psutil
        cpu_temp: CPUTemperature = CPUTemperature()
        while True:
            InfoHolder.temperature = str(cpu_temp.temperature)
            InfoHolder.memory_load = str(psutil.virtual_memory().percent)
            InfoHolder.cpu_load = str(psutil.cpu_percent(interval=0.5))
            time.sleep(0.5)

    def stop(self):
        Shufflecad.stop()
        if not self.is_real_robot:
            self.__connection_helper.stop_channels()
        InfoHolder.logger.write_main_log("Program stopped")

    def handler(self, signum, _):
        InfoHolder.logger.write_main_log("Program stopped")
        InfoHolder.logger.write_main_log('Signal handler called with signal' + str(signum))
        self.stop()
        raise SystemExit("Exited")

    @property
    def motor_speed_0(self):
        return self.__motor_speed_0

    @motor_speed_0.setter
    def motor_speed_0(self, value):
        self.__motor_speed_0 = value
        if not self.is_real_robot:
            self.__update_motors()
        else:
            TitanStatic.speed_motor_0 = value

    @property
    def motor_speed_1(self):
        return self.__motor_speed_1

    @motor_speed_1.setter
    def motor_speed_1(self, value):
        self.__motor_speed_1 = value
        if not self.is_real_robot:
            self.__update_motors()
        else:
            TitanStatic.speed_motor_1 = value

    @property
    def motor_speed_2(self):
        return self.__motor_speed_2

    @motor_speed_2.setter
    def motor_speed_2(self, value):
        self.__motor_speed_2 = value
        if not self.is_real_robot:
            self.__update_motors()
        else:
            TitanStatic.speed_motor_2 = value

    @property
    def motor_speed_3(self):
        return self.__motor_speed_3

    @motor_speed_3.setter
    def motor_speed_3(self, value):
        self.__motor_speed_3 = value
        if not self.is_real_robot:
            self.__update_motors()
        else:
            TitanStatic.speed_motor_3 = value

    @property
    def motor_enc_0(self):
        if not self.is_real_robot:
            self.__update_encs()
            return self.__motor_enc_0
        else:
            return TitanStatic.enc_motor_0

    @property
    def motor_enc_1(self):
        if not self.is_real_robot:
            self.__update_encs()
            return self.__motor_enc_1
        else:
            return TitanStatic.enc_motor_1

    @property
    def motor_enc_2(self):
        if not self.is_real_robot:
            self.__update_encs()
            return self.__motor_enc_2
        else:
            return TitanStatic.enc_motor_2

    @property
    def motor_enc_3(self):
        if not self.is_real_robot:
            self.__update_encs()
            return self.__motor_enc_3
        else:
            return TitanStatic.enc_motor_3

    @property
    def yaw(self):
        if not self.is_real_robot:
            self.__update_sensors()
            return self.__yaw
        else:
            return VMXStatic.yaw

    @property
    def us_1(self):
        if not self.is_real_robot:
            self.__update_sensors()
            return self.__ultrasound_1
        else:
            return VMXStatic.ultrasound_1

    @property
    def us_2(self):
        if not self.is_real_robot:
            self.__update_sensors()
            return self.__ultrasound_2
        else:
            return VMXStatic.ultrasound_2

    @property
    def analog_1(self):
        if not self.is_real_robot:
            self.__update_sensors()
            return self.__analog_1
        else:
            return VMXStatic.analog_1

    @property
    def analog_2(self):
        if not self.is_real_robot:
            self.__update_sensors()
            return self.__analog_2
        else:
            return VMXStatic.analog_2

    @property
    def analog_3(self):
        if not self.is_real_robot:
            self.__update_sensors()
            return self.__analog_3
        else:
            return VMXStatic.analog_3

    @property
    def analog_4(self):
        if not self.is_real_robot:
            self.__update_sensors()
            return self.__analog_4
        else:
            return VMXStatic.analog_4

    @property
    def titan_limits(self) -> list:
        if not self.is_real_robot:
            self.__update_buttons()
            return [self.__limit_h_0, self.__limit_l_0,
                    self.__limit_h_1, self.__limit_l_1,
                    self.__limit_h_2, self.__limit_l_2,
                    self.__limit_h_3, self.__limit_l_3]
        else:
            return [TitanStatic.limit_h_0, TitanStatic.limit_l_0,
                    TitanStatic.limit_h_1, TitanStatic.limit_l_1,
                    TitanStatic.limit_h_2, TitanStatic.limit_l_2,
                    TitanStatic.limit_h_3, TitanStatic.limit_l_3]

    @property
    def vmx_flex(self) -> list:
        if not self.is_real_robot:
            self.__update_buttons()
            return [self.__flex_0, self.__flex_1,
                    self.__flex_2, self.__flex_3,
                    self.__flex_4, self.__flex_5,
                    self.__flex_6, self.__flex_7]
        else:
            return [VMXStatic.flex_0, VMXStatic.flex_1,
                    VMXStatic.flex_2, VMXStatic.flex_3,
                    VMXStatic.flex_4, VMXStatic.flex_5,
                    VMXStatic.flex_6, VMXStatic.flex_7]

    @property
    def camera_image(self):
        if not self.is_real_robot:
            self.__update_camera()
        else:
            try:
                ret, frame = self.__camera_instance.read()
                if ret:
                    self.__camera_image = frame
            except Exception:
                # there could be an error if there is no camera instance
                pass
        return self.__camera_image

    # port is from 1 to 10 included
    def set_angle_hcdio(self, value: float, port: int):
        if not self.is_real_robot:
            dut: float = 0.000666 * value + 0.05
            self.__hcdio_values[port - 1] = dut
            self.__update_oms()
        else:
            VMXStatic.set_servo_angle(value, port - 1)

    # port is from 1 to 10 included
    def set_pwm_hcdio(self, value: float, port: int):
        if not self.is_real_robot:
            self.__hcdio_values[port - 1] = value
            self.__update_oms()
        else:
            VMXStatic.set_servo_pwm(value, port - 1)

    # port is from 1 to 10 included
    def set_bool_hcdio(self, value: bool, port: int):
        if not self.is_real_robot:
            dut: float = 0.2 if value else 0.0
            self.__hcdio_values[port - 1] = dut
            self.__update_oms()
        else:
            VMXStatic.set_led_state(value, port - 1)

    # for virtual robot
    def __update_other(self):
        self.__connection_helper.set_other(
                (
                    0, 0, 0,  # there is no other now
                ))

    def __update_motors(self):
        self.__connection_helper.set_motors(
            (
                self.__motor_speed_0,
                self.__motor_speed_1,
                self.__motor_speed_2,
                self.__motor_speed_3,
            ))

    def __update_oms(self):
        self.__connection_helper.set_oms(tuple(self.__hcdio_values))

    def __update_resets(self):
        self.__connection_helper.set_resets(
            (
                0, 0, 0  # there are no resets now
            ))

    def __update_encs(self):
        values = self.__connection_helper.get_encs()
        if len(values) == 4:
            self.__motor_enc_0 = values[0]
            self.__motor_enc_1 = values[1]
            self.__motor_enc_2 = values[2]
            self.__motor_enc_3 = values[3]

    def __update_sensors(self):
        values = self.__connection_helper.get_sens()
        if len(values) == 7:
            self.__ultrasound_1 = values[0]
            self.__ultrasound_2 = values[1]
            self.__analog_1 = values[2]
            self.__analog_2 = values[3]
            self.__analog_3 = values[4]
            self.__analog_4 = values[5]
            self.__yaw = values[6]

    def __update_buttons(self):
        values = self.__connection_helper.get_buttons()
        if len(values) == 16:
            self.__limit_h_0 = values[0]
            self.__limit_l_0 = values[1]
            self.__limit_h_1 = values[2]
            self.__limit_l_1 = values[3]
            self.__limit_h_2 = values[4]
            self.__limit_l_2 = values[5]
            self.__limit_h_3 = values[6]
            self.__limit_l_3 = values[7]

            self.__flex_0 = values[8]
            self.__flex_1 = values[9]
            self.__flex_2 = values[10]
            self.__flex_3 = values[11]
            self.__flex_4 = values[12]
            self.__flex_5 = values[13]
            self.__flex_6 = values[14]
            self.__flex_7 = values[15]

    def __update_camera(self):
        # because of 640x480
        camera_data = self.__connection_helper.get_camera()
        if len(camera_data) == 921600:
            nparr = np.frombuffer(camera_data, np.uint8)
            if nparr.size > 0:
                img_rgb = nparr.reshape(480, 640, 3)
                img_bgr = cv2.cvtColor(img_rgb, cv2.COLOR_RGB2BGR)
                self.__camera_image = img_bgr



import os
import sys
import time
from threading import Thread

import spidev

from shufflecad.shared import InfoHolder
from .shared import VMXStatic
from funcad.funcad import Funcad


class VMXSPI:
    toggler: int = 0

    @classmethod
    def start_spi(cls) -> None:
        th: Thread = Thread(target=cls.spi_loop)
        th.daemon = True
        th.start()

    @classmethod
    def spi_loop(cls) -> None:
        try:
            spi = spidev.SpiDev()
            spi.open(1, 2)
            spi.max_speed_hz = 1000000
            spi.mode = 0b00

            start_time: float = time.time() * 1000
            send_count_time: float = time.time()
            comm_counter = 0
            while True:
                tx_time: float = time.time() * 1000
                tx_list = cls.set_up_tx_data()
                InfoHolder.tx_spi_time_dev = str(round(time.time() * 1000 - tx_time, 2))

                rx_list = spi.xfer(tx_list)

                rx_time: float = time.time() * 1000
                cls.set_up_rx_data(rx_list)
                InfoHolder.rx_spi_time_dev = str(round(time.time() * 1000 - rx_time, 2))

                comm_counter += 1
                if time.time() - send_count_time > 1:
                    send_count_time = time.time()
                    InfoHolder.spi_count_dev = str(comm_counter)
                    comm_counter = 0

                time.sleep(0.002)
                InfoHolder.spi_time_dev = str(round(time.time() * 1000 - start_time, 2))
                start_time = time.time() * 1000
        except (Exception, EOFError) as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            file_name = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            InfoHolder.logger.write_main_log(" ".join(map(str, [exc_type, file_name, exc_tb.tb_lineno])))
            InfoHolder.logger.write_main_log(str(e))

    @classmethod
    def set_up_rx_data(cls, data: bytearray) -> None:
        if data[0] == 1:
            yaw_ui: int = (data[2] & 0xff) << 8 | (data[1] & 0xff)
            us1_ui: int = (data[4] & 0xff) << 8 | (data[3] & 0xff)
            VMXStatic.ultrasound_1 = us1_ui / 100
            us2_ui: int = (data[6] & 0xff) << 8 | (data[5] & 0xff)
            VMXStatic.ultrasound_2 = us2_ui / 100

            power: float = ((data[8] & 0xff) << 8 | (data[7] & 0xff)) / 100
            InfoHolder.power = str(power)

            # calc yaw unlim
            new_yaw = (yaw_ui / 100) * (1 if Funcad.access_bit(data[9], 1) else -1)
            cls.calc_yaw_unlim(new_yaw, VMXStatic.yaw)
            VMXStatic.yaw = new_yaw

            VMXStatic.flex_0 = Funcad.access_bit(data[9], 2)
            VMXStatic.flex_1 = Funcad.access_bit(data[9], 3)
            VMXStatic.flex_2 = Funcad.access_bit(data[9], 4)
            VMXStatic.flex_3 = Funcad.access_bit(data[9], 5)
            VMXStatic.flex_4 = Funcad.access_bit(data[9], 6)
        elif data[0] == 2:
            VMXStatic.analog_1 = (data[2] & 0xff) << 8 | (data[1] & 0xff)
            VMXStatic.analog_2 = (data[4] & 0xff) << 8 | (data[3] & 0xff)
            VMXStatic.analog_3 = (data[6] & 0xff) << 8 | (data[5] & 0xff)
            VMXStatic.analog_4 = (data[8] & 0xff) << 8 | (data[7] & 0xff)

            VMXStatic.flex_5 = Funcad.access_bit(data[9], 1)
            VMXStatic.flex_6 = Funcad.access_bit(data[9], 2)
            VMXStatic.flex_7 = Funcad.access_bit(data[9], 3)

    @classmethod
    def set_up_tx_data(cls) -> bytearray:
        tx_list: bytearray = bytearray([0x00] * 10)

        if cls.toggler == 0:
            tx_list[0] = 1

            tx_list[9] = 222
        return tx_list

    @staticmethod
    def calc_yaw_unlim(new_yaw: float, old_yaw: float):
        delta_yaw = new_yaw - old_yaw
        if delta_yaw < -180:
            delta_yaw = 180 - old_yaw
            delta_yaw += 180 + new_yaw
        elif delta_yaw > 180:
            delta_yaw = (180 + old_yaw) * -1
            delta_yaw += (180 - new_yaw) * -1
        VMXStatic.yaw_unlim += delta_yaw



import socket
import threading
import time
import warnings

from shufflecad.shared import InfoHolder


class ListenPort:
    def __init__(self, port: int, is_camera=False):
        self.__port = port
        self.__is_camera = is_camera

        # other
        self.__stop_thread = False
        self.out_string = ''
        self.out_bytes = b''

        self.__sct = None
        self.__thread = None

    def start_listening(self):
        self.__thread = threading.Thread(target=self.listening, args=())
        self.__thread.start()

    def listening(self):
        self.__sct = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
        self.__sct.connect(('127.0.0.1', self.__port))
        InfoHolder.logger.write_main_log("connected: " + str(self.__port))
        while not self.__stop_thread:
            try:
                if self.__is_camera:
                    self.__sct.sendall("Wait for size".encode('utf-16-le'))
                    image_size = self.__sct.recv(4)
                    if len(image_size) < 4:
                        continue
                    buffer_size = (image_size[3] & 0xff) << 24 | (image_size[2] & 0xff) << 16 | \
                                  (image_size[1] & 0xff) << 8 | (image_size[0] & 0xff)
                    self.__sct.sendall("Wait for image".encode('utf-16-le'))
                    self.out_bytes = self.__sct.recv(buffer_size)
                else:
                    self.__sct.sendall("Wait for data".encode('utf-16-le'))
                    data_size = self.__sct.recv(4)
                    if len(data_size) < 4:
                        continue
                    length = (data_size[3] & 0xff) << 24 | (data_size[2] & 0xff) << 16 | \
                             (data_size[1] & 0xff) << 8 | (data_size[0] & 0xff)
                    self.out_bytes = self.__sct.recv(length)
                    self.out_string = self.out_bytes.decode('utf-16-le')
                # задержка для слабых компов
                time.sleep(0.004)
            except (ConnectionAbortedError, BrokenPipeError):
                # возникает при отключении сокета
                break
        InfoHolder.logger.write_main_log("disconnected: " + str(self.__port))
        self.__sct.shutdown(socket.SHUT_RDWR)
        self.__sct.close()

    def reset_out(self):
        self.out_string = ''
        self.out_bytes = b''

    def stop_listening(self):
        self.__stop_thread = True
        self.reset_out()
        if self.__sct is not None:
            try:
                self.__sct.shutdown(socket.SHUT_RDWR)
            except (OSError, Exception):
                InfoHolder.logger.write_main_log("Something went wrong while shutting down socket on port " +
                                                 str(self.__port))
            if self.__thread is not None:
                st_time = time.time()
                # если поток все еще живой, ждем 1 секунды и закрываем сокет
                while self.__thread.is_alive():
                    if time.time() - st_time > 1:
                        InfoHolder.logger.write_main_log("Something went wrong. Rude disconnection on port " +
                                                         str(self.__port))
                        try:
                            self.__sct.close()
                        except (OSError, Exception):
                            InfoHolder.logger.write_main_log("Something went wrong while closing socket on port " +
                                                             str(self.__port))
                        st_time = time.time()


class TalkPort:
    def __init__(self, port: int):
        self.__port = port

        # other
        self.__stop_thread = False
        self.out_string = ''

        self.__sct = None
        self.__thread = None

    def start_talking(self):
        self.__thread = threading.Thread(target=self.talking, args=())
        self.__thread.start()

    def talking(self):
        self.__sct = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
        self.__sct.connect(('127.0.0.1', self.__port))
        InfoHolder.logger.write_main_log("connected: " + str(self.__port))
        while not self.__stop_thread:
            try:
                self.__sct.sendall((self.out_string + "$").encode('utf-16-le'))
                _ = self.__sct.recv(1024)  # ответ сервера
                # задержка для слабых компов
                time.sleep(0.004)
            except (ConnectionAbortedError, BrokenPipeError):
                # возникает при отключении сокета
                break
        InfoHolder.logger.write_main_log("disconnected: " + str(self.__port))
        self.__sct.shutdown(socket.SHUT_RDWR)
        self.__sct.close()

    def reset_out(self):
        self.out_string = ''

    def stop_talking(self):
        self.__stop_thread = True
        self.reset_out()
        if self.__sct is not None:
            try:
                self.__sct.shutdown(socket.SHUT_RDWR)
            except (OSError, Exception):
                InfoHolder.logger.write_main_log("Something went wrong while shutting down socket on port " +
                                                 str(self.__port))
            if self.__thread is not None:
                st_time = time.time()
                # если поток все еще живой, ждем 1 секунды и закрываем сокет
                while self.__thread.is_alive():
                    if time.time() - st_time > 1:
                        InfoHolder.logger.write_main_log("Something went wrong. Rude disconnection on port " +
                                                         str(self.__port))
                        try:
                            self.__sct.close()
                        except (OSError, Exception):
                            InfoHolder.logger.write_main_log("Something went wrong while closing socket on port " +
                                                             str(self.__port))
                        st_time = time.time()


class ParseChannels:
    @staticmethod
    def parse_float_channel(txt: str):
        try:
            return tuple(map(float, txt.replace(',', '.').split(';')))
        except (Exception, ValueError):
            return tuple()

    @staticmethod
    def parse_bool_channel(txt: str):
        try:
            return tuple(map(bool, map(int, txt.split(';'))))
        except (Exception, ValueError):
            return tuple()

    @staticmethod
    def join_float_channel(lst: tuple):
        return ';'.join(map(str, lst))

    @staticmethod
    def join_bool_channel(lst: tuple):
        return ';'.join(map(str, map(int, lst)))



import os
import sys
import time
from threading import Thread

import serial

from shufflecad.shared import InfoHolder
from .shared import TitanStatic
from funcad.funcad import Funcad


class TitanCOM:
    @classmethod
    def start_com(cls) -> None:
        th: Thread = Thread(target=cls.com_loop)
        th.daemon = True
        th.start()

    @classmethod
    def com_loop(cls) -> None:
        try:
            ser = serial.Serial(
                port='/dev/ttyACM0',
                baudrate=115200,
                parity=serial.PARITY_NONE,
                stopbits=serial.STOPBITS_ONE,
                bytesize=serial.EIGHTBITS
            )

            start_time: int = round(time.time() * 10000)
            send_count_time: float = time.time()
            comm_counter = 0
            while True:
                rx_data: bytearray = bytearray(ser.read(48))
                ser.reset_input_buffer()  # reset buffer
                rx_time: int = round(time.time() * 10000)
                TitanCOM.set_up_rx_data(rx_data)
                InfoHolder.rx_com_time_dev = str(round(time.time() * 10000) - rx_time)

                tx_time: int = round(time.time() * 10000)
                tx_data = TitanCOM.set_up_tx_data()
                InfoHolder.tx_com_time_dev = str(round(time.time() * 10000) - tx_time)
                ser.reset_output_buffer()  # reset buffer
                ser.write(tx_data)
                ser.flush()

                comm_counter += 1
                if time.time() - send_count_time > 1:
                    send_count_time = time.time()
                    InfoHolder.com_count_dev = str(comm_counter)
                    comm_counter = 0

                time.sleep(0.001)
                InfoHolder.com_time_dev = str(round(time.time() * 10000) - start_time)
                start_time = round(time.time() * 10000)
        except (Exception, serial.SerialException) as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            file_name = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            InfoHolder.logger.write_main_log(" ".join(map(str, [exc_type, file_name, exc_tb.tb_lineno])))
            InfoHolder.logger.write_main_log(str(e))

    @staticmethod
    def set_up_rx_data(data: bytearray) -> None:
        if data[42] != 33:
            if data[0] == 1:
                if data[24] == 111:
                    raw_enc_0: int = (data[2] & 0xff) << 8 | (data[1] & 0xff)
                    raw_enc_1: int = (data[4] & 0xff) << 8 | (data[3] & 0xff)
                    raw_enc_2: int = (data[6] & 0xff) << 8 | (data[5] & 0xff)
                    raw_enc_3: int = (data[8] & 0xff) << 8 | (data[7] & 0xff)
                    TitanCOM.set_up_encoders(raw_enc_0, raw_enc_1, raw_enc_2, raw_enc_3)

                    TitanStatic.limit_l_0 = Funcad.access_bit(data[9], 1)
                    TitanStatic.limit_h_0 = Funcad.access_bit(data[9], 2)
                    TitanStatic.limit_l_1 = Funcad.access_bit(data[9], 3)
                    TitanStatic.limit_h_1 = Funcad.access_bit(data[9], 4)
                    TitanStatic.limit_l_2 = Funcad.access_bit(data[9], 5)
                    TitanStatic.limit_h_2 = Funcad.access_bit(data[9], 6)
                    TitanStatic.limit_l_3 = Funcad.access_bit(data[10], 1)
                    TitanStatic.limit_h_3 = Funcad.access_bit(data[10], 2)
                    # достигнута ли позиция лифта
                    TitanStatic.lift_pos_reached = Funcad.access_bit(data[10], 6)

        else:
            InfoHolder.logger.write_main_log("received wrong data " + " ".join(map(str, data)))

    @staticmethod
    def set_up_tx_data() -> bytearray:
        tx_data: bytearray = bytearray([0] * 48)
        tx_data[0] = 1

        # this cringe was for leds
        tx_data[1] = int('1' + '0000001', 2)

        motor_speeds: bytearray = Funcad.int_to_4_bytes(abs(int(TitanStatic.speed_motor_0 / 100 * 65535)))
        tx_data[2] = motor_speeds[2]
        tx_data[3] = motor_speeds[3]

        motor_speeds: bytearray = Funcad.int_to_4_bytes(abs(int(TitanStatic.speed_motor_1 / 100 * 65535)))
        tx_data[4] = motor_speeds[2]
        tx_data[5] = motor_speeds[3]

        motor_speeds: bytearray = Funcad.int_to_4_bytes(abs(int(TitanStatic.speed_motor_2 / 100 * 65535)))
        tx_data[6] = motor_speeds[2]
        tx_data[7] = motor_speeds[3]

        motor_speeds: bytearray = Funcad.int_to_4_bytes(abs(int(TitanStatic.speed_motor_3 / 100 * 65535)))
        tx_data[8] = motor_speeds[2]
        tx_data[9] = motor_speeds[3]

        tx_data[10] = int('1' + ("1" if TitanStatic.speed_motor_0 >= 0 else "0") +
                          ("1" if TitanStatic.speed_motor_1 >= 0 else "0") +
                          ("1" if TitanStatic.speed_motor_2 >= 0 else "0") +
                          ("1" if TitanStatic.speed_motor_3 >= 0 else "0") + '001', 2)

        # cringe fuck this shite
        tx_data[11] = int('1' + '0100001', 2)

        tx_data[20] = 222

        # if TitanStatic.speed_motor_0 > 0:
        #     Logger.write_com_log("received wrong data " + " ".join(map(str, tx_data)))

        return tx_data

    @staticmethod
    def set_up_encoders(enc_0: int, enc_1: int, enc_2: int, enc_3: int) -> None:
        TitanStatic.enc_motor_0 -= TitanCOM.get_normal_diff(enc_0, TitanStatic.raw_enc_motor_0)
        TitanStatic.enc_motor_1 -= TitanCOM.get_normal_diff(enc_1, TitanStatic.raw_enc_motor_1)
        TitanStatic.enc_motor_2 -= TitanCOM.get_normal_diff(enc_2, TitanStatic.raw_enc_motor_2)
        TitanStatic.enc_motor_3 -= TitanCOM.get_normal_diff(enc_3, TitanStatic.raw_enc_motor_3)

        TitanStatic.raw_enc_motor_0 = enc_0
        TitanStatic.raw_enc_motor_1 = enc_1
        TitanStatic.raw_enc_motor_2 = enc_2
        TitanStatic.raw_enc_motor_3 = enc_3

    @staticmethod
    def get_normal_diff(curr: int, last: int) -> int:
        diff: int = curr - last
        if diff > 30000:
            diff = -(last + (65535 - curr))
        elif diff < -30000:
            diff = curr + (65535 - last)
        return diff



from setuptools import setup, find_packages

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Education',
    'Operating System :: Microsoft :: Windows :: Windows 10',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3'
]

setup(
    name='robocad-py',
    version='0.0.0.6',
    description='python lib for real and virtual robots',
    long_description="Python library for real and virtual robots" + '\n\n' + open('CHANGELOG.md').read(),
    url='https://github.com/CrackAndDie/robocad-py',
    author='Abdrakov Airat',
    author_email='abdrakovairat@gmail.com',
    license='MIT',
    classifiers=classifiers,
    keywords=['simulator', 'robotics', 'robot', '3d', 'raspberry', 'control'],
    packages=find_packages(),
    install_requires=['numpy', 'funcad', 'shufflecad-py']
)



from .connection import TalkPort, ListenPort, ParseChannels


class ConnectionHelper:
    # для connection helper
    CONN_OTHER: int = 0x0001
    CONN_MOTORS_AND_ENCS: int = 0x0002
    CONN_OMS: int = 0x0004
    CONN_RESETS: int = 0x0008
    CONN_SENS: int = 0x0010
    CONN_BUTTONS: int = 0x0020
    CONN_CAMERA: int = 0x0040
    CONN_ALL: int = 0x00ff

    __port_other: int = 65431
    __port_motors: int = 65432
    __port_oms: int = 65433
    __port_resets: int = 65434
    __port_encs: int = 65435
    __port_sens: int = 65436
    __port_buttons: int = 65437
    __port_camera: int = 65438

    def __init__(self, setting: int) -> None:
        self.__setting = setting

        if self.__setting & self.CONN_OTHER:
            self.__other_channel = TalkPort(self.__port_other)
        if self.__setting & self.CONN_MOTORS_AND_ENCS:
            self.__motors_channel = TalkPort(self.__port_motors)
            self.__encs_channel = ListenPort(self.__port_encs)
        if self.__setting & self.CONN_OMS:
            self.__oms_channel = TalkPort(self.__port_oms)
        if self.__setting & self.CONN_RESETS:
            self.__resets_channel = TalkPort(self.__port_resets)
        if self.__setting & self.CONN_SENS:
            self.__sensors_channel = ListenPort(self.__port_sens)
        if self.__setting & self.CONN_BUTTONS:
            self.__buttons_channel = ListenPort(self.__port_buttons)
        if self.__setting & self.CONN_CAMERA:
            self.__camera_channel = ListenPort(self.__port_camera, True)

    def start_channels(self) -> None:
        if self.__setting & self.CONN_OTHER:
            self.__other_channel.start_talking()
        if self.__setting & self.CONN_MOTORS_AND_ENCS:
            self.__motors_channel.start_talking()
            self.__encs_channel.start_listening()
        if self.__setting & self.CONN_OMS:
            self.__oms_channel.start_talking()
        if self.__setting & self.CONN_RESETS:
            self.__resets_channel.start_talking()
        if self.__setting & self.CONN_SENS:
            self.__sensors_channel.start_listening()
        if self.__setting & self.CONN_BUTTONS:
            self.__buttons_channel.start_listening()
        if self.__setting & self.CONN_CAMERA:
            self.__camera_channel.start_listening()

    def stop_channels(self) -> None:
        if self.__setting & self.CONN_OTHER:
            self.__other_channel.stop_talking()
        if self.__setting & self.CONN_MOTORS_AND_ENCS:
            self.__motors_channel.stop_talking()
            self.__encs_channel.stop_listening()
        if self.__setting & self.CONN_OMS:
            self.__oms_channel.stop_talking()
        if self.__setting & self.CONN_RESETS:
            self.__resets_channel.stop_talking()
        if self.__setting & self.CONN_SENS:
            self.__sensors_channel.stop_listening()
        if self.__setting & self.CONN_BUTTONS:
            self.__buttons_channel.stop_listening()
        if self.__setting & self.CONN_CAMERA:
            self.__camera_channel.stop_listening()

    def set_other(self, values: tuple) -> None:
        self.__other_channel.out_string = ParseChannels.join_float_channel(values)

    def set_motors(self, values: tuple) -> None:
        self.__motors_channel.out_string = ParseChannels.join_float_channel(values)

    def set_oms(self, values: tuple) -> None:
        self.__oms_channel.out_string = ParseChannels.join_float_channel(values)

    def set_resets(self, values: tuple) -> None:
        self.__resets_channel.out_string = ParseChannels.join_bool_channel(values)

    def get_encs(self) -> tuple:
        return ParseChannels.parse_float_channel(self.__encs_channel.out_string)

    def get_sens(self) -> tuple:
        return ParseChannels.parse_float_channel(self.__sensors_channel.out_string)

    def get_buttons(self) -> tuple:
        return ParseChannels.parse_bool_channel(self.__buttons_channel.out_string)

    def get_camera(self) -> bytes:
        return self.__camera_channel.out_bytes



import sys


class TitanStatic:
    # from Titan
    speed_motor_0: float = 0.0
    speed_motor_1: float = 0.0
    speed_motor_2: float = 0.0
    speed_motor_3: float = 0.0

    enc_motor_0: int = 0
    enc_motor_1: int = 0
    enc_motor_2: int = 0
    enc_motor_3: int = 0

    raw_enc_motor_0: int = 0
    raw_enc_motor_1: int = 0
    raw_enc_motor_2: int = 0
    raw_enc_motor_3: int = 0

    limit_l_0: bool = False
    limit_h_0: bool = False
    limit_l_1: bool = False
    limit_h_1: bool = False
    limit_l_2: bool = False
    limit_h_2: bool = False
    limit_l_3: bool = False
    limit_h_3: bool = False


class VMXStatic:
    HCDIO_CONST_ARRAY = [4, 18, 17, 27, 23, 22, 24, 25, 7, 5]

    yaw: float = 0
    yaw_unlim: float = 0
    calib_imu: bool = False

    ultrasound_1: float = 0
    ultrasound_2: float = 0

    analog_1: int = 0
    analog_2: int = 0
    analog_3: int = 0
    analog_4: int = 0

    flex_0: bool = False
    flex_1: bool = False
    flex_2: bool = False
    flex_3: bool = False
    flex_4: bool = False
    flex_5: bool = False
    flex_6: bool = False
    flex_7: bool = False

    @classmethod
    def set_servo_angle(cls, angle: float, pin: int):
        dut: float = 0.000666 * angle + 0.05
        VMXStatic.echo_to_file(str(cls.HCDIO_CONST_ARRAY[pin]) + "=" + str(dut))

    @classmethod
    def set_led_state(cls, state: bool, pin: int):
        dut: float = 0.2 if state else 0.0
        VMXStatic.echo_to_file(str(cls.HCDIO_CONST_ARRAY[pin]) + "=" + str(dut))

    @classmethod
    def set_servo_pwm(cls, pwm: float, pin: int):
        dut: float = pwm
        VMXStatic.echo_to_file(str(cls.HCDIO_CONST_ARRAY[pin]) + "=" + str(dut))

    @classmethod
    def disable_servo(cls, pin: int):
        VMXStatic.echo_to_file(str(cls.HCDIO_CONST_ARRAY[pin]) + "=" + "0.0")

    @staticmethod
    def echo_to_file(st: str):
        original_stdout = sys.stdout
        with open('/dev/pi-blaster', 'w') as f:
            sys.stdout = f  # Change the standard output to the file we created.
            print(st)
            sys.stdout = original_stdout  # Reset the standard output to its original value
